#include "viewer.h"
#include "camera.h"
#include "quad.h"


extern int WIDTH;
extern int HEIGHT;

using namespace Eigen;

Viewer::Viewer()
{
}

Viewer::~Viewer()
{
    delete _displayQuad;
}


Vector3f RandomColor() {
    return Vector3f(Eigen::internal::random<float>(0,1),
                    Eigen::internal::random<float>(0,1),
                    Eigen::internal::random<float>(0,1));
}


Affine3f RandomTranslation() {
    return Affine3f(Translation3f(3 * Eigen::internal::random<float>(-1,1),
                                  1.5 * Eigen::internal::random<float>(0,1),
                                  3 * Eigen::internal::random<float>(-1,1)));
}

////////////////////////////////////////////////////////////////////////////////
// GL stuff

// initialize OpenGL context
void Viewer::init(int w, int h){
    _winWidth = w;
    _winHeight = h;

    // Couleur d'arrière plan
    glClearColor(0.0, 0.0, 0.0, 0.0);

    // ballet light initialisation
    for (int i = 0; i < 10; ++i) {
        _pointLight[i] = Sphere(0.2f * Eigen::internal::random<float>(0,1));
        _pointLight[i].init();

        _pointLight[i].setTransformationMatrix(RandomTranslation());

        _lightColor[i] = RandomColor();
    }

    loadProgram();

    _displayQuad = new Quad();
    _displayQuad->init();

    Quad* quad = new Quad();
    quad->init();
    quad->setTransformationMatrix( AngleAxisf((float)M_PI_2, Vector3f(-1,0,0)) * Scaling(20.f,20.f,1.f) * Translation3f(0,0,-0.5));
    _shapes.push_back(quad);
    _specularCoef.push_back(0.);

    Mesh* mesh = new Mesh();
    mesh->load(DATA_DIR"/models/tw.off");
    mesh->init();
    _shapes.push_back(mesh);
    _specularCoef.push_back(0.75);

    mesh = new Mesh();
    mesh->load(DATA_DIR"/models/sphere.off");
    mesh->init();
    mesh->setTransformationMatrix(Translation3f(0,0,2.f) * Scaling(0.5f) );
    _shapes.push_back(mesh);
    _specularCoef.push_back(0.2);


    AlignedBox3f aabb;
    for(int i=0; i<_shapes.size(); ++i)
        aabb.extend(_shapes[i]->boundingBox());

    _cam.setSceneCenter(aabb.center());
    _cam.setSceneRadius(aabb.sizes().maxCoeff());
    _cam.setSceneDistance(_cam.sceneRadius() * 3.f);
    _cam.setMinNear(0.1f);
    _cam.setNearFarOffsets(-_cam.sceneRadius() * 100.f,
                            _cam.sceneRadius() * 100.f);
    _cam.setScreenViewport(AlignedBox2f(Vector2f(0.0,0.0), Vector2f(w,h)));

    glGenSamplers(1, &_samplerId);

    glSamplerParameteri(_samplerId, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glSamplerParameteri(_samplerId, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    glViewport(0, 0, w, h);
    glEnable(GL_DEPTH_TEST);
    _fbo.init(w, h);
}

void Viewer::reshape(int w, int h){
    _winWidth = w;
    _winHeight = h;
    _cam.setScreenViewport(AlignedBox2f(Vector2f(0.0,0.0), Vector2f(w,h)));
    glViewport(0, 0, w, h);
}


/*!
   callback to draw graphic primitives
 */
void Viewer::display()
{
    _fbo.bind();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // update light position

    _gbufferPrg.activate();

    glUniformMatrix4fv(_gbufferPrg.getUniformLocation("projection_matrix"),1,false,_cam.computeProjectionMatrix().data());
    glUniformMatrix4fv(_gbufferPrg.getUniformLocation("view_matrix"),1,false,_cam.computeViewMatrix().data());

    // draw meshes
    for(int i=0; i<_shapes.size(); ++i)
    {
        glUniformMatrix4fv(_gbufferPrg.getUniformLocation("model_matrix"),1,false,_shapes[i]->getTransformationMatrix().data());
        Matrix3f normal_matrix = (_cam.computeViewMatrix()*_shapes[i]->getTransformationMatrix()).linear().inverse().transpose();
        glUniformMatrix3fv(_gbufferPrg.getUniformLocation("normal_matrix"), 1, GL_FALSE, normal_matrix.data());
        glUniform1f(_gbufferPrg.getUniformLocation("specular_coef"),_specularCoef[i]);

        _shapes[i]->display(&_gbufferPrg);
    }
    _gbufferPrg.deactivate();

    //_fbo.savePNG("color");
    //_fbo.savePNG("normal");
    _fbo.unbind();

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glEnable (GL_BLEND);
    glBlendEquation (GL_FUNC_ADD);
    glBlendFunc (GL_ONE, GL_ONE);

    glDisable(GL_DEPTH_TEST);
    for (int i = 0; i < 4; ++i) {
        Vector3f lightPos = (_pointLight[i].getTransformationMatrix()*Vector4f(0,0,0,1)).head(3);
        lightPos[0] += Eigen::internal::random<float>(-1,1) / 10. ;
        lightPos[2] += Eigen::internal::random<float>(-1,1) / 10. ;
        _pointLight[i].setTransformationMatrix(Affine3f(Translation3f(lightPos)));
        Vector4f lightPosH;
        lightPosH << lightPos, 1.f;


        _deferredPrg.activate();

        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, _fbo.textures[0]);
        glUniform1i(_deferredPrg.getUniformLocation("colorBuffer"), 0);
        glBindSampler(0, _samplerId);

        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, _fbo.textures[1]);
        glUniform1i(_deferredPrg.getUniformLocation("normalBuffer"), 1);
        glBindSampler(1, _samplerId);

        glUniformMatrix4fv(_deferredPrg.getUniformLocation("inv_projection_matrix"),1,false,Matrix4f(_cam.computeProjectionMatrix().inverse()).data());
        glUniform4fv(_deferredPrg.getUniformLocation("light_pos"),1, lightPosH.data());
        glUniform3fv(_deferredPrg.getUniformLocation("light_col"),1,_lightColor[i].data());

        glUniform1f(_deferredPrg.getUniformLocation("width"), _winWidth);
        glUniform1f(_deferredPrg.getUniformLocation("height"), _winHeight);
        _displayQuad->display(&_deferredPrg);

        _deferredPrg.deactivate();
    }
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_BLEND);

    for (int i = 0; i < 4; ++i) {
        Vector3f lightPos = (_pointLight[i].getTransformationMatrix()*Vector4f(0,0,0,1)).head(3);
        Vector4f lightPosH;
        lightPosH << lightPos, 1.f;

        _simplePrg.activate();

        glUniformMatrix4fv(_simplePrg.getUniformLocation("projection_matrix"),1,false,_cam.computeProjectionMatrix().data());
        glUniformMatrix4fv(_simplePrg.getUniformLocation("view_matrix"),1,false,_cam.computeViewMatrix().data());
        glUniformMatrix4fv(_simplePrg.getUniformLocation("model_matrix"),1,false,_pointLight[i].getTransformationMatrix().data());
        glUniform4fv(_simplePrg.getUniformLocation("light_pos"),1, lightPosH.data());
        glUniform3fv(_simplePrg.getUniformLocation("light_col"),1,_lightColor[i].data());
        _pointLight[i].display(&_simplePrg);

        _simplePrg.deactivate();
    }
}


void Viewer::updateScene()
{
    display();
}

void Viewer::loadProgram()
{
    _blinnPrg.loadFromFiles(DATA_DIR"/shaders/blinn.vert", DATA_DIR"/shaders/blinn.frag");
    _simplePrg.loadFromFiles(DATA_DIR"/shaders/simple.vert", DATA_DIR"/shaders/simple.frag");
    _gbufferPrg.loadFromFiles(DATA_DIR"/shaders/gbuffer.vert", DATA_DIR"/shaders/gbuffer.frag");
    _deferredPrg.loadFromFiles(DATA_DIR"/shaders/deferred.vert", DATA_DIR"/shaders/deferred.frag");
    checkError();
}

////////////////////////////////////////////////////////////////////////////////
// Events

/*!
   callback to manage mouse : called when user press or release mouse button
   You can change in this function the way the user
   interact with the system.
 */
void Viewer::mousePressed(GLFWwindow *window, int button, int action)
{
    if(action == GLFW_PRESS) {
        if(button == GLFW_MOUSE_BUTTON_LEFT)
        {
            _cam.startRotation(_lastMousePos);
        }
        else if(button == GLFW_MOUSE_BUTTON_RIGHT)
        {
            _cam.startTranslation(_lastMousePos);
        }
        _button = button;
    }else if(action == GLFW_RELEASE) {
        if(_button == GLFW_MOUSE_BUTTON_LEFT)
        {
            _cam.endRotation();
        }
        else if(_button == GLFW_MOUSE_BUTTON_RIGHT)
        {
            _cam.endTranslation();
        }
        _button = -1;
    }
}


/*!
   callback to manage mouse : called when user move mouse with button pressed
   You can change in this function the way the user
   interact with the system.
 */
void Viewer::mouseMoved(int x, int y)
{
    if(_button == GLFW_MOUSE_BUTTON_LEFT)
    {
        _cam.dragRotate(Vector2f(x,y));
    }
    else if(_button == GLFW_MOUSE_BUTTON_RIGHT)
    {
        _cam.dragTranslate(Vector2f(x,y));
    }
    _lastMousePos = Vector2f(x,y);
}

void Viewer::mouseScroll(double x, double y)
{
    _cam.zoom((y>0)? 1.1: 1./1.1);
}

/*!
   callback to manage keyboard interactions
   You can change in this function the way the user
   interact with the system.
 */
void Viewer::keyPressed(int key, int action, int mods)
{
    if(key == GLFW_KEY_R && action == GLFW_PRESS)
        loadProgram();
}

void Viewer::charPressed(int key)
{
}
